import os
import records


class FlaskRecords(records.Database):

    def __init__(self, flask_app, **kwargs):
        self.db_url = flask_app.config['SQLALCHEMY_DATABASE_URI'] or os.environ.get(
            'DATABASE_URL')
        super().__init__(self.db_url, **kwargs)
        flask_app.raw_db = self

    def init_app(self, flask_app, **kwargs):
        self.db_url = flask_app.config['SQLALCHEMY_DATABASE_URI'] or os.environ.get(
            'DATABASE_URL')
        super().__init__(self.db_url, **kwargs)
        flask_app.raw_db = self

    def query_by_page(self, sql, page, page_size, fetchall, params):
        counts = f"SELECT COUNT(*) FROM ({sql}) AS tmp"
        current_page_sql = f"{sql} LIMIT {page_size} OFFSET {(page - 1) * page_size} "
        with self.get_connection() as conn:
            total = conn.query(counts, fetchall, **params).scalar()
            data = conn.query(current_page_sql, fetchall, **params)
            is_last_page = page * page_size >= total
        return data, total, is_last_page



import inspect
from functools import wraps
from collections import namedtuple
from flask import current_app

from .parse_args import get_params_of_func
from .errors import ParameterNotDefinedError


Pagination = namedtuple(
    'Pagination', ['data', 'total', 'is_last_page', 'page', 'page_size'])


def query(sql, fetchall=False):
    def db_query(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            parameters = get_params_of_func(func, *args, **kwargs)
            results = current_app.raw_db.query(sql, fetchall, **parameters)
            return results
        return wrapper
    return db_query


def query_by_page(sql, fetchall=False, page_size=10):
    def db_query(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            parameters = get_params_of_func(func, *args, **kwargs)
            page = parameters.get('page', None)
            if not page:
                raise ParameterNotDefinedError('page')

            data, total, is_last_page = current_app.raw_db.query_by_page(
                sql, int(page), int(page_size), fetchall, parameters)

            pagination = Pagination._make(
                [data, total, is_last_page, page, page_size])
            return pagination
        return wrapper
    return db_query



import unittest
from tests.app import app, db, BasicTestCase
from flask_records.decorators import query, query_by_page
from flask_records import RecordsDao


@app.route('/dao/pagination/static/user/r/<int:page>')
def static_query_user(page):
    pagination = UserDao.query_user_by_page_with_static_method(page)
    return pagination.data[0].name


@app.route('/dao/pagination/cls/user/r/<int:page>')
def cls_query_user(page):
    pagination = UserDao().query_user_by_page_with_cls_method(page)
    return pagination.data[0].name


@app.route('/dao/pagination/func/user/r/<int:page>')
def function_query_user(page):
    pagination = query_user_by_page_with_function(page)
    return pagination.data[0].name


class UserDao(RecordsDao):

    def __init__(self):
        super(UserDao, self).__init__()

    @staticmethod
    @query_by_page('SELECT * FROM user', True, page_size=2)
    def query_user_by_page_with_static_method(page):
        pass

    @query_by_page('SELECT * FROM user', True, page_size=2)
    def query_user_by_page_with_cls_method(self, page):
        pass


@query_by_page('SELECT * FROM user', True, page_size=2)
def query_user_by_page_with_function(page):
    pass


class PaginationTestCase(BasicTestCase):

    def test_1_query_by_page_with_static_method(self):
        self.app.get('/dao/user/c/xc/1')
        self.app.get('/dao/user/c/my/1')
        self.app.get('/dao/user/c/leo/18')

        response = self.app.get('/dao/pagination/static/user/r/1')
        assert b'xc' == response.data

        response = self.app.get('/dao/pagination/static/user/r/2')
        assert b'leo' == response.data

    def test_2_query_by_page_with_cls_method(self):

        response = self.app.get('/dao/pagination/cls/user/r/1')
        assert b'xc' == response.data

        response = self.app.get('/dao/pagination/cls/user/r/2')
        assert b'leo' == response.data

    def test_3_query_by_page_with_function(self):
        response = self.app.get('/dao/pagination/func/user/r/1')
        assert b'xc' == response.data

        response = self.app.get('/dao/pagination/func/user/r/2')
        assert b'leo' == response.data


if __name__ == '__main__':
    unittest.main()



import unittest
from tests.app import app, db, BasicTestCase
from flask_records.decorators import query, query_by_page


@app.route('/basic/user/c/<name>/<age>')
def create_user(name, age):
    basic_create_user(name, age)
    return 'OK'


@app.route('/basic/user/r/<id>')
def read_user(id):
    user = get_user_by_id(id).first()
    if not user:
        return 'The user does not exist!'
    else:
        return user.name


@app.route('/basic/user/u/<id>/<name>')
def update_user(id, name='my'):
    update_user_by_id(id, name)
    return 'OK'


@app.route('/basic/user/d/<id>')
def delete_user(id):
    delete_user_by_id(id)
    return 'OK'


@query("INSERT INTO user(name, age) VALUES (:name, :age)")
def basic_create_user(name, age):
    pass


@query("SELECT * FROM user WHERE id = :id", True)
def get_user_by_id(id):
    pass


@query("UPDATE user SET name = :new_name WHERE id = :id")
def update_user_by_id(id, new_name):
    pass


@query("DELETE FROM user where id = :id")
def delete_user_by_id(id):
    pass


class TestCase(BasicTestCase):

    def test_1_create(self):
        # create user xc with 1 age
        response = self.app.get('/basic/user/c/xc/1')
        assert b'OK' == response.data

    def test_2_read(self):
        response = self.app.get('/basic/user/r/1')
        assert b'xc' == response.data

    def test_3_update(self):
        self.app.get('/basic/user/u/1/my')
        response = self.app.get('/basic/user/r/1')
        assert b'my' == response.data

    def test_4_delete(self):
        self.app.get('/basic/user/d/1')
        response = self.app.get('/basic/user/r/1')
        self.assertEqual(b'The user does not exist!', response.data)


if __name__ == '__main__':
    unittest.main()



import inspect
from types import MethodType


def is_func_or_static_method_in_decorator(func, *args):
    """
    determine if the func is the real func, static method or class method of the class

    If no args, it should be function or static method of the class

    if there are arguments, get the first argument and 
    try to validate if it contains the func.__name__ or not,
    if yes,  determine it with isinstance(attr, MethodType)
    else it should be a function or static method
    """
    is_function_or_static_method = True
    if len(args) > 0 and hasattr(args[0], func.__name__):
        attr = getattr(args[0], func.__name__)
        if isinstance(attr, MethodType):
            is_function_or_static_method = False

    return is_function_or_static_method


def get_params_of_func(func, *args, **kwargs):
    """
    retrieve the parameters of the func, and organize as a dict to pass to the db engine
    """
    is_function_or_static_method = is_func_or_static_method_in_decorator(
        func, *args)
    signatures = list(inspect.signature(func).parameters.keys())
    if not is_function_or_static_method:
        signatures.pop(0)
        _, *args = args
    for signature, value in zip(signatures, args):
        if isinstance(value, dict):
            kwargs.update(value)
        else:
            kwargs.update({signature: value})
    return kwargs



class Error(Exception):
    pass


class ParameterNotDefinedError(Error):

    def __init__(self, parameters):
        self.message = f'{parameters} is required'


class NotFoundFieldError(Error):

    def __init__(self, parameters, table):
        self.message = f'{parameters} is not in the {table} table'



import unittest
from tests.app import app, db, BasicTestCase
from flask_records.decorators import query, query_by_page
from flask_records import RecordsDao


# @app.route('/dao/user/c/<name>/<age>')
# def dao_create_user(name, age):
#     UserDao().create({'name': name, 'age': age})
#     return 'OK'


@app.route('/dao/pagination/user/r/<int:page>')
def query_user(page):
    pagination = UserDao().query_user_by_page(page)
    return pagination.data[0].name


class UserDao(RecordsDao):

    def __init__(self):
        super(UserDao, self).__init__()

    @query_by_page('SELECT * FROM user', True, page_size=2)
    def query_user_by_page(self, page):
        pass


class PaginationTestCase(BasicTestCase):

    def test_query_by_page(self):
        self.app.get('/dao/user/c/xc/1')
        self.app.get('/dao/user/c/my/1')
        self.app.get('/dao/user/c/leo/18')

        response = self.app.get('/dao/pagination/user/r/1')
        assert b'xc' == response.data

        response = self.app.get('/dao/pagination/user/r/2')
        assert b'leo' == response.data


if __name__ == '__main__':
    unittest.main()



import unittest
from tests.app import app, db, BasicTestCase
from flask_records.decorators import query, query_by_page
from flask_records import RecordsDao


@app.route('/dao/user/c/<name>/<age>')
def dao_create_user(name, age):
    UserDao().create({'name': name, 'age': 1})
    return 'OK'


@app.route('/dao/user/r/<int:id>')
def dao_read_user(id):
    user = UserDao().get(id).first()
    if not user:
        return 'The user does not exist!'
    else:
        return user.name


@app.route('/dao/user/u/<id>/<name>')
def dao_update_user(id, name):
    UserDao().update({'id': id, 'name': 'my'})
    return 'OK'


@app.route('/dao/user/d/<id>')
def dao_delete_user(id):
    UserDao().delete(id)
    return 'OK'


class UserDao(RecordsDao):

    def __init__(self):
        super(UserDao, self).__init__()


class DaoTestCase(BasicTestCase):

    def test_1_create_dao(self):
        response = self.app.get('/dao/user/c/xc/1')
        assert b'OK' == response.data

    def test_2_read_dao(self):
        response = self.app.get('/dao/user/r/1')
        assert b'xc' == response.data

    def test_3_update_dao(self):
        self.app.get('/dao/user/u/1/my')
        response = self.app.get('/dao/user/r/1')
        assert b'my' == response.data

    def test_4_delete_dao(self):
        self.app.get('/dao/user/d/1')
        response = self.app.get('/dao/user/r/1')

        self.assertEqual(b'The user does not exist!', response.data)


if __name__ == '__main__':
    unittest.main()



import os
import unittest

from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from flask_records import FlaskRecords
from flask_records.decorators import query, query_by_page
from flask_records.records_dao import RecordsDao

basedir = os.path.dirname(__file__)

app = Flask(__name__)
app.config['SQLALCHEMY_DATABASE_URI'] = f'sqlite:////{basedir}/app.db'
app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

db = SQLAlchemy(app)
raw_db = FlaskRecords(app)
# raw_db.init_app(app)


class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(128))
    age = db.Column(db.Integer)


class BasicTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.app = app.test_client()
        db.create_all()

    @classmethod
    def tearDownClass(cls):
        # pass
        db.session.remove()
        db.drop_all()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="flask_records",
    version="0.0.8",
    author="leo",
    author_email="leo.anonymous@qq.com",
    description="Flask wrapper for the SQL Records",
    # 包的详细介绍(一般通过加载README.md)
    long_description=long_description,
    # 和上条命令配合使用，声明加载的是markdown文件
    long_description_content_type="text/markdown",
    # 项目开源地址，我这里写的是同性交友官网，大家可以写自己真实的开源网址
    url="https://github.com/",
    # 如果项目由多个文件组成，我们可以使用find_packages()自动发现所有包和子包，而不是手动列出每个包，在这种情况下，包列表将是example_pkg
    packages=setuptools.find_packages(),
    classifiers=[                                           # 关于包的其他元数据(metadata)
        "Programming Language :: Python :: 3",              # 该软件包仅与Python3兼容
        "License :: OSI Approved :: MIT License",           # 根据MIT许可证开源
        "Operating System :: OS Independent",               # 与操作系统无关
    ],
    install_requires=['Flask>=0.9',
                      'Flask-SQLAlchemy>=1.0',
                      'records>=0.5.3'],
    test_suite="tests"
)



import re
from sqlalchemy import inspect
from flask import current_app

from functools import wraps
from .decorators import query
from .errors import NotFoundFieldError


class RecordsDao:
    """
    The Basic Dao which provide the simple crud operations as default.
    """

    CREATE_SQL = 'INSERT INTO {}({}) VALUES({})'
    DELETE_SQL = 'DELETE FROM {} WHERE id = :id'
    UPDATE_SQL = 'UPDATE {} SET {} WHERE id = :id'
    SELECT_SQL = 'SELECT * FROM {} WHERE id = :id'

    def __init__(self):
        self.table_name = self._get_table_name()

    def create(self, attributes):
        """
        Create a new table record 
        """
        columns = list(attributes.keys())

        # will prevent the malicious code through this function
        self._check_dynamic_columns_in_raw_sql(columns)

        values = []
        for column in columns:
            values.append(f':{column}')

        @query(RecordsDao.CREATE_SQL.format(self.table_name, ','.join(columns), ','.join(values)))
        def _create(attributes):
            pass

        return _create(attributes)

    def delete(self, id):
        """
        Delete the record according to the id
        """
        @query(RecordsDao.DELETE_SQL.format(self.table_name))
        def _delete(id):
            pass

        return _delete(id)

    def update(self, attributes):
        """
        Update the record according to the id, the id need be contained in the attributes
        """
        columns = list(attributes.keys())
        columns.remove('id')
        # will prevent the malicious code through this function
        self._check_dynamic_columns_in_raw_sql(columns)

        column_assembly = []
        for column in columns:
            column_assembly.append(f'{column} = :{column}')

        @query(RecordsDao.UPDATE_SQL.format(self.table_name, ','.join(column_assembly)))
        def _update(attributes):
            pass

        return _update(attributes)

    def get(self, id):
        """
        Get the record according to the id
        """
        @query(RecordsDao.SELECT_SQL.format(self.table_name), True)
        def _get(id):
            pass

        return _get(id)

    def _check_dynamic_columns_in_raw_sql(self, column_parameters):
        """
        Check the columns is in db table or not.

        if there is the column which pass from client side is not in the table,
        raise the Exception to prevent in case the malicious code is executed.
        """
        db_table_columns = self._get_table_columns()
        difference = [
            column for column in column_parameters if column not in db_table_columns]
        if len(difference) > 0:
            raise NotFoundFieldError(','.join(difference), self.table_name)

    def _get_table_columns(self):
        """
        Refect the DB engine to get the metadata of the table
        """
        inspector = inspect(current_app.raw_db._engine)

        # solve the conflict on the user table
        _table_name = self._class_name_to_table_name_transformation()
        if _table_name == 'user':
            columns = inspector.get_columns(_table_name)
        else:
            columns = inspector.get_columns(self.table_name)

        columns = [column['name'] for column in columns]

        return columns

    def _get_table_name(self):
        """
        Get the table name according to the class name which should be written with CamelCase style.

        if the __tablename__ is specified, will use it as the table name else translate class name
        to table_name
        """
        try:
            table_name = self.__tablename__
        except AttributeError:
            table_name = self._class_name_to_table_name_transformation()
        except Exception:
            raise Exception('Failed to get the table_name')
        return table_name

    def _class_name_to_table_name_transformation(self):
        """
        tanslate the class name to the table name according to the named conventions

        1. split the class name with the capital letter

        2. if 'Dao' is included in the splitted array, will remove it, then join the elements
        in the array with '_' sign, and lower the string you get

        For example:
            CamelCase => camel_case
            CamelCaseDao => camel_case
            Camel => camel
        """
        name_array = re.split('([A-Z][a-z0-9]*)', self.__class__.__name__)
        exclude_key_words = ('', 'Dao')
        table_name_array = [
            name for name in name_array if name not in exclude_key_words]
        table_name = '_'.join(table_name_array).lower()
        return table_name
